"""
Useful definitions used throughout the IDR.
"""

from drapi.code.drapi.drapi import flatExtend

__all__ = ["LIST_OF_PHI_VARIABLES_BO",
           "LIST_OF_PHI_VARIABLES_CLINICAL_TEXT",
           "LIST_OF_PHI_VARIABLES_OMOP",
           "VARIABLE_NAME_TO_FILE_NAME_DICT",
           "FILE_NAME_TO_VARIABLE_NAME_DICT"]

DICT_OF_PHI_VARIABLES_BIRTHDATES = {"BO": ["Date Of Birth"],
                                    "OMOP": ["birth_datetime"]}
DICT_OF_PHI_VARIABLES_AGE = {"BO": ["Current Age",
                                    "Age At Diagnosis",
                                    "Age at Encounter"]}

LIST_OF_PHI_DATES_BO = ["Date of Diagnosis",
                        "Chemo Start Date Summary",
                        "Combined Last Contact",
                        "Date Of Birth",
                        "Date Of Conclusive Diagnosis",
                        "Date Of Diagnosis",
                        "Date Of Multiple Tumors",
                        "Date Of Sentinel Lymph Node Biopsy",
                        "Date Regional Lymph Node Dissection",
                        "Date Systemic Therapy Started",
                        "Date of Diagnosis - Year",
                        "First Contact Date",
                        "First Recurrence Date",
                        "Hormone Start Date",
                        "Immuno Start Date",
                        "Initial Rx Date",
                        "Most Definitive Surgery Date",
                        "Non Cancer Directed Surgery Date",
                        "Other Rx Start Date",
                        "Radiation End Date",
                        "Radiation Start Date",
                        "Surgery Date Summary",
                        "Surgery Discharge Date Summary"]

LIST_OF_PHI_DATES_CLINICAL_TEXT = ["ContactDate",
                                   "CreatedDatetime",
                                   "EncounterDate",
                                   "OrderPlacedDatetime",
                                   "OrderResultDatetime",
                                   "ServiceDatetime"]

LIST_OF_PHI_DATES_OMOP = ["birth_datetime",
                          "condition_end_date",
                          "condition_end_datetime",
                          "condition_start_date",
                          "condition_start_datetime",
                          "death_date",
                          "death_datetime",
                          "device_exposure_end_date",
                          "device_exposure_end_datetime",
                          "device_exposure_start_date",
                          "device_exposure_start_datetime",
                          "drug_exposure_end_date",
                          "drug_exposure_end_datetime",
                          "drug_exposure_start_date",
                          "drug_exposure_start_datetime",
                          "measurement_date",
                          "measurement_datetime",
                          "observation_date",
                          "observation_datetime",
                          "observation_period_end_date",
                          "observation_period_start_date",
                          "procedure_date",
                          "procedure_datetime",
                          "verbatim_end_date",
                          "visit_end_date",
                          "visit_end_datetime",
                          "visit_start_date",
                          "visit_start_datetime"]


LIST_OF_PHI_VARIABLES_BO = ["Acct Number - Enter DateTime Comb",
                            "Acct Number - Exit DateTime Comb",
                            "At Station",
                            "Authoring Provider Key",
                            "Authorizing Provider Key",
                            "Chemotherapy Rx Hosp Code Desc",
                            "Cosign Provider Key",
                            "Diagnosis County",
                            "EPIC Patient ID",
                            "Encounter #",
                            "Encounter # (CSN)",
                            "Encounter # (Primary CSN)",
                            "Encounter Key",
                            "Encounter Key (Primary CSN)",
                            "Enterprise ID",
                            "From Station",
                            "F/u Physicians",
                            "Linkage Note ID",
                            "Location of Svc",
                            "Location of Svc ID",
                            "Medical Record Number",
                            "Managing Physician",
                            "MRN (Jax)",
                            "MRN (UF)",
                            "NRAS",  # This is PHI but cannot be de-identified, so it should be deleted
                            "Note ID",
                            "Note Key",
                            "Order ID",
                            "Ordering Provider Key",
                            "Order Key",
                            "Prim  surgeon Code",  # NOTE: Double space is intentional
                            "Provider Key",
                            "Patient Encounter Key",
                            "Patient Key",
                            "Patnt Key",
                            "Radiation Hosp Code Desc",
                            "Source Sys",
                            "Surgery Rx Hosp Code Desc",
                            "To Station"]

LIST_OF_PHI_VARIABLES_I2B2 = ["LOCATION_CD"]

LIST_OF_PHI_VARIABLES_CLINICAL_TEXT = ["AuthoringProviderKey",
                                       "AuthorizingProviderKey",
                                       "CosignProviderKey",
                                       "EncounterCSN",
                                       "EncounterKey",
                                       "LinkageNoteID",
                                       "MRN_GNV",
                                       "MRN_JAX",
                                       "NoteID",
                                       "NoteKey",
                                       "OrderID",
                                       "OrderKey",
                                       "OrderingProviderKey",
                                       "PatientKey",
                                       "Patient Key",
                                       "ProviderKey"]

# The list defined `LIST_OF_PHI_VARIABLES_OMOP_BIRTHDATE` is provided for reference. Each variable on its own is not PHI, but in combination with others they are PHI.
LIST_OF_PHI_VARIABLES_OMOP_BIRTHDATE_CONDITIONAL = ["year_of_birth",
                                                    "month_of_birth",
                                                    "day_of_birth"]

# The list `LIST_OF_PHI_VARIABLES_OMOP_UNINFORMATIVE` contains variables that are not analytically valuable because they only uniquely identify the row in a table.
LIST_OF_PHI_VARIABLES_OMOP_UNINFORMATIVE = ["condition_occurrence_id",
                                            "device_exposure_id",
                                            "drug_exposure_id",
                                            "location_source_value",
                                            "measurement_id",
                                            "observation_id",
                                            "observation_period_id",
                                            "procedure_occurrence_id",
                                            "visit_detail_id"]

LIST_OF_PHI_VARIABLES_OMOP = ["care_site_id",
                              "csn",
                              "city",
                              "county",
                              "location_id",
                              "patient_key",
                              "preceding_visit_occurrence_id",
                              "provider_id",
                              "visit_occurrence_id"]

LIST_OF_PHI_VARIABLES = flatExtend([LIST_OF_PHI_VARIABLES_BO,
                                    LIST_OF_PHI_VARIABLES_I2B2,
                                    LIST_OF_PHI_VARIABLES_CLINICAL_TEXT,
                                    LIST_OF_PHI_VARIABLES_OMOP])

# Variable suffixes: By portion
VARIABLE_SUFFIXES_BO = {"Authoring Provider Key": {"columnSuffix": "provider",
                                                   "deIdIDSuffix": "PROV"},
                        "Authorizing Provider Key": {"columnSuffix": "provider",
                                                     "deIdIDSuffix": "PROV"},
                        "Accct Number - Enter DateTime Comb": {"columnSuffix": "account",
                                                               "deIdIDSuffix": "ACCT"},
                        "Acct Number - Exit DateTime Comb": {"columnSuffix": "account",
                                                             "deIdIDSuffix": "ACCT"},
                        "At Station": {"columnSuffix": "station",
                                       "deIdIDSuffix": "STN"},
                        "Chemotherapy Rx Hosp Code Desc": {"columnSuffix": "location",
                                                           "deIdIDSuffix": "LOC"},
                        "Cosign Provider Key": {"columnSuffix": "provider",
                                                "deIdIDSuffix": "PROV"},
                        "Diagnosis County": {"columnSuffix": "location",
                                             "deIdIDSuffix": "LOC"},
                        "EPIC Patient ID": {"columnSuffix": "patient",
                                            "deIdIDSuffix": "PAT"},
                        "Encounter #": {"columnSuffix": "encounter",
                                        "deIdIDSuffix": "ENC"},
                        "Encounter Key": {"columnSuffix": "encounter",
                                          "deIdIDSuffix": "ENC"},
                        "EncounterCSN": {"columnSuffix": "encounter",
                                         "deIdIDSuffix": "ENC"},
                        "Encounter # (CSN)": {"columnSuffix": "encounter",
                                              "deIdIDSuffix": "ENC"},
                        "Encounter # (Primary CSN)": {"columnSuffix": "encounter",
                                                      "deIdIDSuffix": "ENC"},
                        "Encounter Key (Primary CSN)": {"columnSuffix": "encounter",
                                                        "deIdIDSuffix": "ENC"},
                        "Enterprise ID": {"columnSuffix": "patient",
                                          "deIdIDSuffix": "PAT"},
                        "F/u Physicians": {"columnSuffix": "provider",
                                           "deIdIDSuffix": "PROV"},
                        "From Station": {"columnSuffix": "station",
                                         "deIdIDSuffix": "STN"},
                        "Linkage Note ID": {"columnSuffix": "link_note",
                                            "deIdIDSuffix": "LINK_NOTE"},
                        "Location of Svc": {"columnSuffix": "location",
                                            "deIdIDSuffix": "LOC"},
                        "Location of Svc ID": {"columnSuffix": "location",
                                               "deIdIDSuffix": "LOC"},
                        "Medical Record Number": {"columnSuffix": "patient",
                                                  "deIdIDSuffix": "PAT"},
                        "Managing Physician": {"columnSuffix": "provider",
                                               "deIdIDSuffix": "PROV"},
                        "MRN (Jax)": {"columnSuffix": "patient",
                                      "deIdIDSuffix": "PAT"},
                        "MRN (UF)": {"columnSuffix": "patient",
                                     "deIdIDSuffix": "PAT"},
                        "NRAS": {"columnSuffix": "NRAS",
                                 "deIdIDSuffix": "NRAS"},
                        "Note ID": {"columnSuffix": "note",
                                    "deIdIDSuffix": "NOTE"},
                        "Note Key": {"columnSuffix": "note",
                                     "deIdIDSuffix": "NOTE"},
                        "Order ID": {"columnSuffix": "order",
                                     "deIdIDSuffix": "ORD"},
                        "Order Key": {"columnSuffix": "order",
                                      "deIdIDSuffix": "ORD"},
                        "Ordering Provider Key": {"columnSuffix": "order",
                                                  "deIdIDSuffix": "ORD"},
                        "Patient Encounter Key": {"columnSuffix": "encounter",
                                                  "deIdIDSuffix": "ENC"},
                        "Patient Key": {"columnSuffix": "patient",
                                        "deIdIDSuffix": "PAT"},
                        "Patnt Key": {"columnSuffix": "patient",
                                      "deIdIDSuffix": "PAT"},
                        "Prim  surgeon Code": {"columnSuffix": "provider",
                                               "deIdIDSuffix": "PROV"},
                        "Provider Key": {"columnSuffix": "provider",
                                         "deIdIDSuffix": "PROV"},
                        "Radiation Hosp Code Desc": {"columnSuffix": "location",
                                                     "deIdIDSuffix": "LOC"},
                        "Source Sys": {"columnSuffix": "location",
                                       "deIdIDSuffix": "LOC"},
                        "Surgery Rx Hosp Code Desc": {"columnSuffix": "location",
                                                      "deIdIDSuffix": "LOC"},
                        "To Station": {"columnSuffix": "station",
                                       "deIdIDSuffix": "STN"}}

VARIABLE_SUFFIXES_CLINICAL_TEXT = {"AuthoringProviderKey": {"columnSuffix": "provider",
                                                            "deIdIDSuffix": "PROV"},
                                   "AuthorizingProviderKey": {"columnSuffix": "provider",
                                                              "deIdIDSuffix": "PROV"},
                                   "CosignProviderKey": {"columnSuffix": "provider",
                                                         "deIdIDSuffix": "PROV"},
                                   "EncounterCSN": {"columnSuffix": "encounter",
                                                    "deIdIDSuffix": "ENC"},
                                   "EncounterKey": {"columnSuffix": "encounter",
                                                    "deIdIDSuffix": "ENC"},
                                   "LinkageNoteID": {"columnSuffix": "link_note",
                                                     "deIdIDSuffix": "LINK_NOTE"},
                                   "MRN_GNV": {"columnSuffix": "patient",
                                               "deIdIDSuffix": "PAT"},
                                   "MRN_JAX": {"columnSuffix": "patient",
                                               "deIdIDSuffix": "PAT"},
                                   "NoteID": {"columnSuffix": "note",
                                              "deIdIDSuffix": "NOTE"},
                                   "NoteKey": {"columnSuffix": "note",
                                               "deIdIDSuffix": "NOTE"},
                                   "OrderID": {"columnSuffix": "order",
                                               "deIdIDSuffix": "ORD"},
                                   "OrderKey": {"columnSuffix": "order",
                                                "deIdIDSuffix": "ORD"},
                                   "OrderingProviderKey": {"columnSuffix": "order",
                                                           "deIdIDSuffix": "ORD"},
                                   "PatientKey": {"columnSuffix": "patient",
                                                  "deIdIDSuffix": "PAT"},
                                   "ProviderKey": {"columnSuffix": "provider",
                                                   "deIdIDSuffix": "PROV"}}

VARIABLE_SUFFIXES_I2B2 = {"LOCATION_CD": {"columnSuffix": "location",
                                          "deIdIDSuffix": "LOC"}}

VARIABLE_SUFFIXES_OMOP = {"care_site_id": {"columnSuffix": "location",
                                           "deIdIDSuffix": "LOC"},
                          "city": {"columnSuffix": "location",
                                   "deIdIDSuffix": "LOC"},
                          "county": {"columnSuffix": "location",
                                     "deIdIDSuffix": "LOC"},
                          "csn": {"columnSuffix": "encounter",
                                  "deIdIDSuffix": "ENC"},
                          "location_id": {"columnSuffix": "location",
                                          "deIdIDSuffix": "LOC"},
                          "observation_period_id": {"columnSuffix": "encounter",
                                                    "deIdIDSuffix": "ENC"},
                          "patient_key": {"columnSuffix": "patient",
                                          "deIdIDSuffix": "PAT"},
                          "person_id": {"columnSuffix": "patient",
                                        "deIdIDSuffix": "PAT"},
                          "preceding_visit_occurrence_id": {"columnSuffix": "encounter",
                                                            "deIdIDSuffix": "ENC"},
                          "provider_id": {"columnSuffix": "provider",
                                          "deIdIDSuffix": "PROV"},
                          "visit_occurrence_id": {"columnSuffix": "encounter",
                                                  "deIdIDSuffix": "ENC"}}
# Variable suffixes: All
VARIABLE_SUFFIXES_LIST = [VARIABLE_SUFFIXES_BO,
                          VARIABLE_SUFFIXES_CLINICAL_TEXT,
                          VARIABLE_SUFFIXES_I2B2,
                          VARIABLE_SUFFIXES_OMOP]
VARIABLE_SUFFIXES = dict()
for variableSuffixDict in VARIABLE_SUFFIXES_LIST:
    VARIABLE_SUFFIXES.update(variableSuffixDict)

# Variable name to file name map. This is necessary because some variable names have characters which are not allowed in file names, e.g., "/".
VARIABLE_NAME_TO_FILE_NAME_DICT = {varName: varName for varName in LIST_OF_PHI_VARIABLES}

# Over-write the necessary values
VARIABLES_TO_OVER_WRITE = {"F/u Physicians": "F-u Physicians"}
VARIABLE_NAME_TO_FILE_NAME_DICT.update(VARIABLES_TO_OVER_WRITE)
FILE_NAME_TO_VARIABLE_NAME_DICT = {fileName: varName for varName, fileName in VARIABLE_NAME_TO_FILE_NAME_DICT.items()}

# QA
for key in VARIABLE_SUFFIXES_BO.keys():
    checkBO = key not in list(VARIABLE_SUFFIXES_CLINICAL_TEXT.keys()) + list(VARIABLE_SUFFIXES_OMOP.keys())
for key in VARIABLE_SUFFIXES_CLINICAL_TEXT.keys():
    checkClinicalText = key not in list(VARIABLE_SUFFIXES_BO.keys()) + list(VARIABLE_SUFFIXES_OMOP.keys())
for key in VARIABLE_SUFFIXES_OMOP.keys():
    checkOMOP = key not in list(VARIABLE_SUFFIXES_BO.keys()) + list(VARIABLE_SUFFIXES_CLINICAL_TEXT.keys())

assert all([checkBO, checkClinicalText, checkOMOP]), "Some variables are present in more than one variable suffix dictionary. This may lead to unintended consequences."



"""
Useful definitions used throughout IDR

See the Clinical Text Portion for current IDR mapping standards: /Volumes/FILES/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Bian/IRB202202436/Intermediate Results/Clinical Text Portion/Data/Output/mapping
See Noah's data request for my attempt at using these standards: /Volumes/FILES/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Bian/IRB202202436/Concatenated Results/Code/makeMap.py
"""
from drapi.code.drapi.constants.phiVariables import LIST_OF_PHI_VARIABLES

__all__ = ["DeIdIDName2DeIdIDSuffix",
           "IDName2DeIdIDName",
           "mapDtypes",
           "DATA_TYPES_DICT"]

IDName2DeIdIDNameRoot = {"ENCNTR_CSN_ID": "enc",
                         "IDENT_ID_INT": "pat",
                         "ORDR_PROC_ID": "order",
                         "PATNT_KEY": "pat"}
IDName2DeIdIDName = {IDName: f"deid_{DeIdIDNameRoot}_id" for IDName, DeIdIDNameRoot in IDName2DeIdIDNameRoot.items()}
DeIdIDName2DeIdIDSuffix = {"pat": "PAT",
                           "enc": "ENC",
                           "note": "NOTE",
                           "link_note": "LINK_NOTE",
                           "order": "ORDER",
                           "provider": "PROV"}
mapDtypes = {0: int,
             1: int,
             2: str}

# De-identification prefixes
DE_IDENTIFICATION_PREFIXES = {"classic": "deid",
                              "DRAPI-Lemur": "De-identified"}

# Define data types. NOTE that all the `String` variables contain numbers with leading zeros that get converted to integers in the current process. Either we convert these variables to `string` or we change the process.
DATA_TYPES_BO = {"Acct Number - Enter DateTime Comb": "String",
                 "Acct Number - Exit DateTime Comb": "String",
                 "At Station": "String",
                 "Authoring Provider Key": "Numeric",
                 "Authorizing Provider Key": "Numeric",
                 "Chemotherapy Rx Hosp Code Desc": "String",
                 "Cosign Provider Key": "Numeric",
                 "Diagnosis County": "String",
                 "EPIC Patient ID": "String",
                 "Encounter #": "Numeric",
                 "Encounter # (CSN)": "Numeric",
                 "Encounter # (Primary CSN)": "Numeric",
                 "Encounter Key": "Numeric",
                 "Encounter Key (Primary CSN)": "Numeric",
                 "EncounterCSN": "Numeric",
                 "Enterprise ID": "String",
                 "From Station": "String",
                 "F/u Physicians": "String",
                 "Linkage Note ID": "Numeric",
                 "Location of Svc": "String",
                 "Location of Svc ID": "String",
                 "Managing Physician": "String",
                 "Medical Record Number": "Numeric",
                 "MRN (Jax)": "Numeric",
                 "MRN (UF)": "Numeric",
                 "Note ID": "Numeric",
                 "Note Key": "Numeric",
                 "NRAS": "String",
                 "Order ID": "Numeric",
                 "Ordering Provider Key": "Numeric",
                 "Order Key": "Numeric",
                 "Provider Key": "Numeric",
                 "Patient Encounter Key": "Numeric",
                 "Patient Key": "Numeric",
                 "PatientKey": "Numeric",
                 "Patnt Key": "Numeric",
                 "Prim  surgeon Code": "String",
                 "Radiation Hosp Code Desc": "String",
                 "Source Sys": "String",
                 "Surgery Rx Hosp Code Desc": "String",
                 "To Station": "String"}

# Note that notes data are from the same source as BO data. These variable names are actually aliases and are here for convenience.
DATA_TYPES_I2B2 = {"LOCATION_CD": "Numeric"}
DATA_TYPES_NOTES = {"AuthoringProviderKey": "Numeric",
                    "AuthorizingProviderKey": "Numeric",
                    "ContactDate": "Datetime",
                    "CosignProviderKey": "Numeric",
                    "CreatedDatetime": "Datetime",
                    "EncounterCSN": "Numeric",
                    "EncounterDate": "Datetime",
                    "EncounterKey": "Numeric",
                    "LinkageNoteID": "Numeric",
                    "MRN_GNV": "Numeric",
                    "MRN_JAX": "Numeric",
                    "NoteID": "Numeric",
                    "NoteKey": "Numeric",
                    "OrderID": "Numeric",
                    "OrderKey": "Numeric",
                    "OrderPlacedDatetime": "Datetime",
                    "OrderResultDatetime": "Datetime",
                    "OrderingProviderKey": "Numeric",
                    "PatientKey": "Numeric",
                    "ProviderKey": "Numeric",
                    "ServiceDatetime": "Datetime"}

DATA_TYPES_OMOP = {"care_site_id": "Numeric",
                   "csn": "Numeric",
                   "city": "String",
                   "county": "String",
                   "location_id": "Numeric",
                   "observation_period_id": "Numeric",
                   "patient_key": "Numeric",
                   "person_id": "Numeric",
                   "preceding_visit_occurrence_id": "Numeric",
                   "provider_id": "Numeric",
                   "visit_occurrence_id": "Numeric"}

DATA_TYPES_DICT = DATA_TYPES_BO.copy()
DATA_TYPES_DICT.update(DATA_TYPES_I2B2)
DATA_TYPES_DICT.update(DATA_TYPES_NOTES)
DATA_TYPES_DICT.update(DATA_TYPES_OMOP)

DATA_TYPES_BY_PORTION = {"BO": DATA_TYPES_BO,
                         "Clinical Text": DATA_TYPES_NOTES,
                         "OMOP": DATA_TYPES_OMOP}

# Convert DRAPI data types to SQL data types
DRAPI_TO_SQL_DATA_TYPES_MAP = {"Datetime": "TEXT",
                               "Numeric": "INTEGER",
                               "String": "TEXT"}
DATA_TYPES_DICT_SQL = {name: DRAPI_TO_SQL_DATA_TYPES_MAP[drapiDataType] for name, drapiDataType in DATA_TYPES_DICT.items()}

# QA: Make sure all PHI variables have their data type defined
# assert all([varName in DATA_TYPES_DICT.keys() for varName in LIST_OF_PHI_VARIABLES])
listOfMissingVariables = []
for varName in LIST_OF_PHI_VARIABLES:
    if varName not in DATA_TYPES_DICT.keys():
        listOfMissingVariables.append(varName)
if len(listOfMissingVariables) > 0:
    text = "\n".join([f'"{varName}"' for varName in listOfMissingVariables])
    raise Exception(f"Not all PHI variables have their data type defined: {text}")



"""
Copy files to a single destination directory. Optionally this directory can be added to a compressed archive.
"""

import logging
import os
import pprint
import shutil
import zipfile
from pathlib import Path
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
LIST_OF_DIRECTORIES = ["../../Concatenated Results/disclosure/Portion Symlinks/Clinical Text",  # Clinical Text
                       "../../Concatenated Results/disclosure/Portion Symlinks/Clinical Text Metadata",  # Clinical Text Metadata
                       "../../Concatenated Results/disclosure/Portion Symlinks/OMOP"]  # OMOP
LIST_OF_DIRECTORIES_NEW_NAMES = ["Clinical Text",
                                 "Clinical Text Metadata",
                                 "OMOP"]
LIST_OF_FILES = []
TIMESTAMP = getTimestamp()
DESTINATION_FOLDER = fr"disclosure/{TIMESTAMP}"

OVERWRITE_IF_EXISTS_FOLDER = False
OVERWRITE_IF_EXISTS_FILE = False

CREATE_MERGED_FOLDER = False
CREATE_COMPRESSED_ARCHIVE = True
DELETE_FOLDER_AFTER_ARCHIVING = False

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 4

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `LIST_OF_DIRECTORIES`: "{LIST_OF_DIRECTORIES}"
    `LIST_OF_DIRECTORIES_NEW_NAMES`: "{LIST_OF_DIRECTORIES_NEW_NAMES}"
    `LIST_OF_FILES`: "{LIST_OF_FILES}"
    `DESTINATION_FOLDER`: "{DESTINATION_FOLDER}"

    `OVERWRITE_IF_EXISTS_FOLDER`: "{OVERWRITE_IF_EXISTS_FOLDER}"
    `OVERWRITE_IF_EXISTS_FILE`: "{OVERWRITE_IF_EXISTS_FILE}"

    `CREATE_COMPRESSED_ARCHIVE`: "{CREATE_COMPRESSED_ARCHIVE}"
    `DELETE_FOLDER_AFTER_ARCHIVING`: "{DELETE_FOLDER_AFTER_ARCHIVING}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Define the destination path
    destinationRootFolder = Path(DESTINATION_FOLDER)

    # Copy files and directories
    if CREATE_MERGED_FOLDER:
        if destinationRootFolder.exists():
            logger.warning(f"""WARNING: The destination folder already exists: "{destinationRootFolder.absolute()}".""")
            if OVERWRITE_IF_EXISTS_FOLDER:
                logger.info("""  Removing folder contents to make room for new files.""")
                for fpath in destinationRootFolder.iterdir():
                    logger.info(f"""    Removing "{fpath.absolute()}".""")
                    os.remove(fpath)
                    logger.info(f"""    Removing "{fpath.absolute()}" - done.""")
            else:
                msg = "  The destination folder exists and no option was passed to over-write it."
                logger.fatal(msg)
                raise Exception(msg)
        else:
            logger.info(f"""Making destination folder: "{destinationRootFolder.absolute()}".""")
            makeDirPath(destinationRootFolder)

        logger.info("""Working on list of files.""")
        for fpathString in LIST_OF_FILES:
            fpath = Path(fpathString)
            logger.info(f"""  Working on file "{fpath.absolute()}".""")
            dest = destinationRootFolder.joinpath(fpath.name)
            logger.info(f"""    Saving to "{dest.absolute()}".""")
            shutil.copyfile(fpath, dest)
        logger.info("""Working on list of files - done.""")

        logger.info("""Working on list of directories.""")
        if LIST_OF_DIRECTORIES_NEW_NAMES:
            listOfDirectoriesNewNames = LIST_OF_DIRECTORIES_NEW_NAMES[:]
        else:
            listOfDirectoriesNewNames = ["" for _ in LIST_OF_DIRECTORIES]
        directoryPathsAndNames = zip(LIST_OF_DIRECTORIES, listOfDirectoriesNewNames)
        for directoryString, newDirectoryName in directoryPathsAndNames:
            directory = Path(directoryString)
            destinationFolder = destinationRootFolder.joinpath(newDirectoryName)
            destinationFolder.mkdir()
            listOfFiles1 = sorted(list(directory.iterdir()))
            for fpath in listOfFiles1:
                logger.info(f"""  Working on file "{fpath.absolute()}".""")
                dest = destinationFolder.joinpath(fpath.name)
                logger.info(f"""    The destination path is "{dest.absolute()}".""")
                if dest.exists():
                    msg = f"""    WARNING: This file already exists: "{dest}"."""
                    logger.warning(msg)
                    if OVERWRITE_IF_EXISTS_FILE:
                        continueOperation = True
                    else:
                        continueOperation = False
                else:
                    continueOperation = True
                if continueOperation:
                    logger.info("""    Saving to destination path.""")
                    shutil.copyfile(fpath, dest)
                else:
                    logger.info("""    The file was not saved to the destination path. File over-write is set to `False`.""")
        logger.info("""Working on list of directories - done.""")

    # Create compressed archive
    if CREATE_COMPRESSED_ARCHIVE:
        archivePath = destinationRootFolder.with_suffix(".ZIP")
        logger.info(f"""Creating compressed archive: "{archivePath.absolute()}"".""")
        if archivePath.exists():
            logger.info("""The archive folder already exists and will be removed before writing.""")
            os.remove(archivePath)
        else:
            destinationRootFolder.mkdir(parents=True)

        # Define list of files to archive, and their paths
        if LIST_OF_DIRECTORIES_NEW_NAMES:
            listOfDirectoriesNewNames = LIST_OF_DIRECTORIES_NEW_NAMES[:]
        else:
            listOfDirectoriesNewNames = ["" for _ in LIST_OF_DIRECTORIES]
        directoryPathsAndNames = zip(LIST_OF_DIRECTORIES, listOfDirectoriesNewNames)
        listOfFiles2All = []
        for directoryString, newDirectoryName in directoryPathsAndNames:
            directory = Path(directoryString)
            listOfFiles2 = sorted(list(directory.iterdir()))
            for fpath2 in listOfFiles2:
                dest2 = Path(TIMESTAMP).joinpath(newDirectoryName, fpath2.name)
                listOfFiles2All.append((fpath2, dest2))
        
        logger.info(f"""These are the files paths for the archived items:\n{pprint.pformat(listOfFiles2All)}.""")

        with zipfile.ZipFile(file=archivePath,
                             mode="a",
                             compression=zipfile.ZIP_DEFLATED) as zipObj:
            logger.info("Adding files to archive.")
            for fpath2, dest2 in sorted(listOfFiles2All):
                logger.info(f"""  Working on file "{fpath2.absolute()}".""")
                logger.info(f"""    This has a destination path of "{dest2}".""")
                zipObj.write(filename=fpath2, arcname=dest2)
        logger.info("Creating compressed archive - done.")

        if CREATE_MERGED_FOLDER:
            if DELETE_FOLDER_AFTER_ARCHIVING:
                logger.info("""Removing intermediate folder.""")
                shutil.rmtree(destinationRootFolder)
                logger.info("""Removing intermediate folder - done.""")
            else:
                pass

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute()}".""")



"""
Aliases for variables.
"""

VAR_ALIASES_CLINICAL_TEXT_ENCOUNTERS = {"EncounterCSN": "Encounter # (CSN)",
                                        "EncounterKey": "Encounter Key"}
VAR_ALIASES_CLINICAL_TEXT_NOTES = {"NoteID": "Note ID",
                                   "NoteKey": "Note Key",
                                   "LinkageNoteID": "Linkage Note ID"}
VAR_ALIASES_CLINICAL_TEXT_ORDERS = {"OrderID": "Order ID",
                                    "OrderKey": "Order Key"}
VAR_ALIASES_CLINICAL_TEXT_PATIENTS = {"MRN_GNV": "MRN (UF)",
                                      "MRN_JAX": "MRN (Jax)",
                                      "PatientKey": "Patient Key"}
VAR_ALIASES_CLINICAL_TEXT_PROVIDERS = {"AuthoringProviderKey": "Provider Key",
                                       "AuthorizingProviderKey": "Provider Key",
                                       "CosignProviderKey": "Provider Key",
                                       "OrderingProviderKey": "Provider Key",
                                       "ProviderKey": "Provider Key"}
VAR_ALIASES_SDOH_ENCOUNTERS = {"NOTE_ENCNTR_KEY": "Linkage Note ID",
                               "NOTE_KEY": "Note Key"}
LIST_OF_ALIAS_DICTS = [VAR_ALIASES_CLINICAL_TEXT_ENCOUNTERS,
                       VAR_ALIASES_CLINICAL_TEXT_NOTES,
                       VAR_ALIASES_CLINICAL_TEXT_ORDERS,
                       VAR_ALIASES_CLINICAL_TEXT_PATIENTS,
                       VAR_ALIASES_CLINICAL_TEXT_PROVIDERS,
                       VAR_ALIASES_SDOH_ENCOUNTERS]
VARIABLE_ALIASES = {}
for di in LIST_OF_ALIAS_DICTS:
    VARIABLE_ALIASES.update(di)




"""
Convert person ID column to patient key

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""

import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    successiveParents)
from drapi.code.drapi.idealist.idealist import idealistMap2dict
# Local packages: Script parameters: General
from common import DATA_REQUEST_ROOT_DIRECTORY_DEPTH
# Local packages: Script parameters: Paths
from common import OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN
# Local packages: Script parameters: File criteria
from common import OMOP_PORTION_FILE_CRITERIA

# Arguments
COLUMNS_TO_CONVERT_DI = {"person_id": "Patient Key"}

MAC_PATHS = [OMOP_PORTION_DIR_MAC]
WIN_PATHS = [OMOP_PORTION_DIR_WIN]

LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA]

PERSON_ID_MAP_PATH = Path(r"..\Intermediate Results\OMOP Portion\data\output\getPersonIDs\...\personIDsFound.csv")  # TODO
CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Load person ID map
    personIDMap = pd.read_csv(PERSON_ID_MAP_PATH)
    personIDMapDi = idealistMap2dict(personIDMap, "person_id", "Patient Key")

    # Convert columns
    logging.info("""Getting the set of values for each variable to convert.""")
    # Create file parameters dictionary

    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logging.info(f"""Working on directory "{directory.relative_to(IRBDir)}".""")
        for file in directory.iterdir():
            logging.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logging.info("""    File has met all conditions for processing.""")
                numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE)])
                dfChunks = readDataFile(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    logging.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        logging.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in COLUMNS_TO_CONVERT_DI.keys():
                            logging.info("""  ..  ..  Column must be converted. Converting values.""")
                            dfChunk[columnName] = dfChunk[columnName].apply(lambda IDNum: personIDMapDi[IDNum])
                            dfChunk = dfChunk.rename(columns={columnName: COLUMNS_TO_CONVERT_DI[columnName]})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logging.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logging.info("""    This file does not need to be processed.""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Iterates over the files that would be processed in the pipeline and runs quality tests on them. Currently the quality tests are

  1. Check for delimmeter issues. This is done by reading the whole file with `pd.read_csv`. Usually if there's an unexpected presence or absence of a delimmeter this will raise an error.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""


# Third-party packages
import pandas as pd
from pandas.errors import ParserError
# Local packages
from drapi.drapi import getTimestamp, makeDirPath


def dataQualityTest(listOfPortionDirs,
                    LIST_OF_PORTION_CONDITIONS,
                    CHUNK_SIZE,
                    pipelineOutputDir,
                    logger,
                    ROOT_DIRECTORY,
                    rootDirectory):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    `listOfPortionDirs`: "{listOfPortionDirs}"
    `LIST_OF_PORTION_CONDITIONS`: "{LIST_OF_PORTION_CONDITIONS}"
    `CHUNK_SIZE`: "{CHUNK_SIZE}"
    `pipelineOutputDir`: "{pipelineOutputDir}"
    `logger`: "{logger}"
    `ROOT_DIRECTORY`: "{ROOT_DIRECTORY}"
    `rootDirectory`: "{rootDirectory}"
    """)

    # Data quality check
    logger.info("""Getting the set of values for each variable to de-identify.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Read file
                logger.info("""    This file has met all conditions for testing.""")
                # Test 1: Make sure all lines have the same number of delimiters
                logger.info("""  ..  Test 1: Make sure all lines have the same number of delimiters.""")
                try:
                    for it, _ in enumerate(pd.read_csv(file, chunksize=CHUNK_SIZE), start=1):
                        logger.info(f"""  ..    Working on chunk {it}...""")
                    logger.info("""  ..    There are no apparent problems reading this file.""")
                except ParserError as err:
                    msg = err.args[0]
                    logger.info(f"""  ..    This file raised an error: "{msg}".""")
                # Test 2: ...
                pass
            else:
                logger.info("""    This file does not need to be tested.""")

    # Return path to sets fo ID values
    # TODO If this is implemented as a function, instead of a stand-alone script, return `runOutputDir` to define `setsPathDir` in the "makeMap" scripts.
    logger.info(f"""Finished performing the battery of tests. Results, if any, will be located the run output directory: "{runOutputDir.relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")



"""
Script for downloading data using SQL queries.
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
from typing import Union
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.getData.getData import getData


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main: Multiple query option
    parser.add_argument("--DICTIONARY_OF_ARGUMENTS",
                        type=json.loads)

    # Arguments: Main: Single query option
    parser.add_argument("--CONNECTION_STRING")
    parser.add_argument("--SQL_FILE_PATH",
                        type=str)
    parser.add_argument("--OUTPUT_FILE_NAME",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_FILE_PATH",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_COLUMN_NAME",
                        type=str)
    if False:
        parser.add_argument("--FILTER_VARIABLE_DATA")  # Not implemented
    parser.add_argument("--FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_PYTHON_DATA_TYPE",
                        type=str,
                        choices=["int", "str"],
                        help="""This affects how the data is sorted and if it's quoted or not in the SQL query.""")
    parser.add_argument("--FILTER_VARIABLE_CHUNK_SIZE",
                        default=10000,
                        type=int)
    parser.add_argument("--QUERY_CHUNK_SIZE",
                        default=10000,
                        type=int)

    parser.add_argument("SCRIPT_TEST_MODE",
                        type=lambda stringValue: True if stringValue.lower() == "true" else False if stringValue.lower() == "false" else None,
                        help=""" Choose one of {{True, False}}""")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main: Multiple query option
    DICTIONARY_OF_ARGUMENTS = argNamespace.DICTIONARY_OF_MAPPING_ARGUMENTS

    # Parsed arguments: Main: Single query option
    CONNECTION_STRING = argNamespace.CONNECTION_STRING
    SQL_FILE_PATH = argNamespace.SQL_FILE_PATH
    OUTPUT_FILE_NAME = argNamespace.OUTPUT_FILE_NAME
    FILTER_VARIABLE_FILE_PATH = argNamespace.FILTER_VARIABLE_FILE_PATH
    FILTER_VARIABLE_COLUMN_NAME = argNamespace.FILTER_VARIABLE_COLUMN_NAME
    FILTER_VARIABLE_DATA = argNamespace.FILTER_VARIABLE_DATA
    FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER = argNamespace.FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER
    FILTER_VARIABLE_PYTHON_DATA_TYPE = argNamespace.FILTER_VARIABLE_PYTHON_DATA_TYPE
    FILTER_VARIABLE_CHUNK_SIZE = argNamespace.FILTER_VARIABLE_CHUNK_SIZE
    QUERY_CHUNK_SIZE = argNamespace.QUERY_CHUNK_SIZE

    # Parsed arguments: Main
    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # Argument parsing: Additional checks  # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    SINGLE_OPTION_ARGUMENTS = [CONNECTION_STRING,
                               SQL_FILE_PATH,
                               OUTPUT_FILE_NAME,
                               FILTER_VARIABLE_FILE_PATH,
                               FILTER_VARIABLE_COLUMN_NAME,
                               FILTER_VARIABLE_DATA,
                               FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                               FILTER_VARIABLE_PYTHON_DATA_TYPE,
                               FILTER_VARIABLE_CHUNK_SIZE,
                               QUERY_CHUNK_SIZE]
    if DICTIONARY_OF_ARGUMENTS and any(SINGLE_OPTION_ARGUMENTS):
        parser.error("""This program is meant to function one of two ways. Either
1. Pass `DICTIONARY_OF_ARGUMENTS`, or
2. Pass each of the single-option arguments
    2. 1. `CONNECTION_STRING`
    2. 2. `SQL_FILE_PATH`
    2. 3. `OUTPUT_FILE_NAME`
    2. 4. `FILTER_VARIABLE_FILE_PATH`
    2. 5. `FILTER_VARIABLE_COLUMN_NAME`
    2. 6. `FILTER_VARIABLE_DATA`
    2. 7. `FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER`
    2. 8. `FILTER_VARIABLE_PYTHON_DATA_TYPE`
    2. 9. `FILTER_VARIABLE_CHUNK_SIZE`
    2. 10. `QUERY_CHUNK_SIZE`""")

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Meta
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    # Conform mapping arguments
    if DICTIONARY_OF_ARGUMENTS:
        sqlFileSettings = DICTIONARY_OF_ARGUMENTS
    elif all(SINGLE_OPTION_ARGUMENTS):
        sqlFileSettings = {0: {"CONNECTION_STRING": CONNECTION_STRING,
                               "FILTER_VARIABLE_FILE_PATH": FILTER_VARIABLE_FILE_PATH,
                               "FILTER_VARIABLE_COLUMN_NAME": FILTER_VARIABLE_COLUMN_NAME,
                               "FILTER_VARIABLE_DATA": FILTER_VARIABLE_DATA,
                               "FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER": FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                               "FILTER_VARIABLE_PYTHON_DATA_TYPE": FILTER_VARIABLE_PYTHON_DATA_TYPE,
                               "FILTER_VARIABLE_CHUNK_SIZE": FILTER_VARIABLE_CHUNK_SIZE,
                               "OUTPUT_FILE_NAME": OUTPUT_FILE_NAME,
                               "QUERY_CHUNK_SIZE": QUERY_CHUNK_SIZE,
                               "SQL_FILE_PATH": SQL_FILE_PATH}}
    else:
        raise Exception("This should not happen")

    # Iterate over SQL file settings
    logger.info("""Iterating over SQL directory contents.""")
    sqlFiles = sorted([(key, di["SQL_FILE_PATH"]) for key, di in sqlFileSettings.items()], key=lambda tu: tu[1])
    for dataKey, sqlFilePath in sqlFiles:
        logger.info(f"""  Working on data key "{dataKey}" with SQL file path "{sqlFilePath}".""")
        fileSettings = sqlFileSettings[dataKey]
        SQL_FILE_PATH = fileSettings["SQL_FILE_PATH"]
        CONNECTION_STRING = fileSettings["CONNECTION_STRING"]
        FILTER_VARIABLE_CHUNK_SIZE = fileSettings["FILTER_VARIABLE_CHUNK_SIZE"]
        FILTER_VARIABLE_COLUMN_NAME = fileSettings["FILTER_VARIABLE_COLUMN_NAME"]
        FILTER_VARIABLE_DATA = fileSettings["FILTER_VARIABLE_DATA"]
        FILTER_VARIABLE_FILE_PATH = fileSettings["FILTER_VARIABLE_FILE_PATH"]
        FILTER_VARIABLE_PYTHON_DATA_TYPE = fileSettings["FILTER_VARIABLE_PYTHON_DATA_TYPE"]
        FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER = fileSettings["FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER"]
        OUTPUT_FILE_NAME = fileSettings["OUTPUT_FILE_NAME"]
        QUERY_CHUNK_SIZE = fileSettings["QUERY_CHUNK_SIZE"]
        getData(sqlFilePath=SQL_FILE_PATH,
                connectionString=CONNECTION_STRING,
                filterVariableChunkSize=FILTER_VARIABLE_CHUNK_SIZE,
                filterVariableColumnName=FILTER_VARIABLE_COLUMN_NAME,
                filterVariableData=FILTER_VARIABLE_DATA,
                filterVariableFilePath=FILTER_VARIABLE_FILE_PATH,
                filterVariablePythonDataType=FILTER_VARIABLE_PYTHON_DATA_TYPE,
                filterVariableSqlQueryTemplatePlaceholder=FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                outputFileName=OUTPUT_FILE_NAME,
                queryChunkSize=QUERY_CHUNK_SIZE,
                runOutputDir=runOutputDir,
                logger=logger)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Processes the notes text TSV files in preparation for processing by the SDOH pipeline.
"""

import argparse
import logging
import os
import pprint
from pathlib import Path
from typing import Union
# Third-party packages
import pandas as pd
# Local packages
from drapi.drapi import getTimestamp, successiveParents, makeDirPath

# Arguments
_ = None

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = 4
IRB_DIR_DEPTH = 4
IDR_DATA_REQUEST_DIR_DEPTH = 6

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter(f"""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

# Functions

def modifyTSV(fpath: Union[str, Path]):
    """
    Modifies notes text TSV files in preparation for analysis in the SDOH pipeline.
    Modifications:
        - Adds two dummy columns
        - Re-orders the columns
        - Renames the `OrderKey` column to "ordr_proc_key"
    """
    # Read file
    df = pd.read_csv(fpath, delimiter="\t")

    # Set dummy values
    df["CNTCT_NUM"] = 0
    df["NOTE_KEY"] = 123456789

    # Re-order columns
    df = df[["LinkageNoteID",
             "NOTE_KEY",
             "CNTCT_NUM",
             "note_text"]]
    
    # Rename columns
    df = df.rename(columns={"LinkageNoteID": "NOTE_ENCNTR_KEY"})

    # Save modified TSV file
    fname = fpath.name
    savePath = runOutputDir.joinpath(fname)
    df.to_csv(savePath, index=False, sep="\t")


if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory.absolute()}".""")

    # Parse arguments
    parser = argparse.ArgumentParser(description="Configure")

    parser.add_argument("--verbosity",
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""",
                        type=int,
                        default=10)

    parser.add_argument("fpath",
                        help="The path(s) to the TSV file(s).",
                        type=str,
                        nargs="+")

    args = parser.parse_args()

    fpathli = args.fpath
    verbosity = args.verbosity

    argsDict = vars(args)

    # Log arguments
    argumentsText = pprint.pformat(argsDict)
    logger.info(f"""Script arguments:

    # Arguments: Command-line arguments
    {argumentsText}

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"
    """)

    # Execute commands
    logger.setLevel(verbosity)
    for fpath in fpathli:
        fpath = Path(fpath)
        logger.info(f"""  Working on file "{fpath.absolute().relative_to(rootDirectory)}".""")
        # modifyTSV(fpath)

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Converts i2b2 `PATIENT_NUM` and `ENCOUNTER_NUM` IDs to "Patient Key" and "Encounter # (CSN)", respectively.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
I2B2_PORTION_OUTPUT_DIR_PATH = Path(r"..\Intermediate Results\i2b2 Portion\data\output\i2b2_dump\...\i2b2")
I2B2_COHORT_IDS_FILE_PATH = Path(r"..\Intermediate Results\i2b2 Portion\data\Cohort IDs.CSV")
ENCOUNTER_MAP_PATH = Path(r"..\Concatenated Results\data\output\i2b2MakeMap\...\i2b2 Encounter Map.CSV")
PATIENT_MAP_PATH1 = Path(r"..\Concatenated Results\data\output\i2b2MakeMap\...\i2b2 Patient Map.CSV")  # i2b2 to EPIC Patient ID
PATIENT_MAP_PATH2 = Path(r"..\Concatenated Results\data\output\i2b2MakeMap\...\i2b2 Patient Map.CSV")  # EPIC Patient ID to Patient Key

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"
connection = create_engine(conStr).connect().execution_options(stream_results=True)

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    `I2B2_PORTION_OUTPUT_DIR_PATH`: "{I2B2_PORTION_OUTPUT_DIR_PATH}"
    `I2B2_COHORT_IDS_FILE_PATH`: "{I2B2_COHORT_IDS_FILE_PATH}"
    `ENCOUNTER_MAP_PATH`: "{ENCOUNTER_MAP_PATH}"
    `PATIENT_MAP_PATH1`: "{PATIENT_MAP_PATH1}"
    `PATIENT_MAP_PATH2`: "{PATIENT_MAP_PATH2}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load maps
    logger.info("""Loading maps.""")
    encounterMap = pd.read_csv(ENCOUNTER_MAP_PATH)

    # Get un-linked patients and drop them.
    cohortIDs = pd.read_csv(I2B2_COHORT_IDS_FILE_PATH)
    unlinkedPatientsMask = cohortIDs["Patient Key"].isna()
    unlinkedPatients = cohortIDs["I2B2_PATIENT_NUM"][unlinkedPatientsMask]

    TARGET_ID_NAME = "Patient Key"

    colnamedict = {"ENCOUNTER_NUM": encounterMap.columns[1],
                   "PATIENT_NUM": TARGET_ID_NAME}

    # Convert maps to dictionaries
    logger.info("""Converting maps to dictionaries.""")
    encounterDict = {k: v for k, v in zip(encounterMap.iloc[:, 0], encounterMap.iloc[:, 1])}
    encounterDict[""] = ""
    patientDict3 = {k: v for k, v in zip(cohortIDs["I2B2_PATIENT_NUM"], cohortIDs[TARGET_ID_NAME])}

    # Convert i2b2 IDs
    logger.info("""Converting i2b2 IDs.""")
    MESSAGE_MODULO = 50
    for fpath in I2B2_PORTION_OUTPUT_DIR_PATH.iterdir():
        logger.info(f"""  Working on file "{fpath.absolute().relative_to(rootDirectory)}".""")
        if fpath.suffix.lower() == ".csv":
            numChunks = 0
            CHUNKSIZE = 10000
            for _ in pd.read_csv(fpath, chunksize=CHUNKSIZE):
                numChunks += 1
            # Logger block
            if numChunks < MESSAGE_MODULO:
                numChunksTenth = numChunks
            else:
                numChunksTenth = round(numChunks / MESSAGE_MODULO)

            # Iterate over file chunks
            header = True
            mode = "w"
            savepath = runOutputDir.joinpath(fpath.name)
            for it, chunk in enumerate(pd.read_csv(fpath, chunksize=CHUNKSIZE), start=1):
                if it % numChunksTenth == 0:
                    logger.info(f"""    Working on chunk {it} of {numChunks}.""")
                chunk = pd.DataFrame(chunk)
                newChunk = chunk.copy()

                # Drop un-linked patients
                mdrop = newChunk["PATIENT_NUM"].isin(unlinkedPatients)
                newChunk = newChunk[~mdrop]

                # Convert patient IDs
                if "PATIENT_NUM" in chunk.columns:
                    newChunk["PATIENT_NUM"] = newChunk["PATIENT_NUM"].apply(lambda i2b2ID: patientDict3[i2b2ID])
                    newChunk = newChunk.rename(columns={"PATIENT_NUM": colnamedict["PATIENT_NUM"]})

                # Convert encounter IDs
                if "ENCOUNTER_NUM" in chunk.columns:
                    newChunk = newChunk.copy()
                    mask = newChunk["ENCOUNTER_NUM"] < 0
                    newChunk.loc[mask, "ENCOUNTER_NUM"] = ""
                    newChunk["ENCOUNTER_NUM"] = newChunk["ENCOUNTER_NUM"].apply(lambda i2b2ID: encounterDict[i2b2ID])
                    newChunk = newChunk.rename(columns={"ENCOUNTER_NUM": colnamedict["ENCOUNTER_NUM"]})

                # Save converted chunk
                newChunk.to_csv(savepath, header=header, mode=mode, index=False)
                header = False
                mode = "a"
        else:
            pass

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Get the set of ID values for all variables to de-identify.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
"""

# Third-party packages
import pandas as pd
import pprint  # noqa
# Local packages
from drapi.drapi import getTimestamp, makeDirPath


def getIDValues(SETS_PATH,
                COLUMNS_TO_DE_IDENTIFY,
                VARIABLE_ALIASES,
                listOfPortionDirs,
                LIST_OF_PORTION_CONDITIONS,
                CHUNK_SIZE,
                ROOT_DIRECTORY,
                rootDirectory,
                pipelineOutputDir,
                logger):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Get set of values
    if SETS_PATH:
        logger.info(f"""Using the set of values previously collected from "{SETS_PATH}".""")
    else:
        logger.info("""Getting the set of values for each variable to de-identify.""")
        mapNames = [columnName for columnName in COLUMNS_TO_DE_IDENTIFY if columnName not in VARIABLE_ALIASES.keys()]
        columnSetsVarsDi = {columnName: {"fpath": runOutputDir.joinpath(f"{columnName}.txt"),
                                         "fileMode": "w"} for columnName in mapNames}
        for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
            # Act on directory
            logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
            for file in directory.iterdir():
                logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
                conditions = [condition(file) for condition in fileConditions]
                if all(conditions):
                    # Read file
                    logger.info("""    File has met all conditions for processing.""")
                    numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                    dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                    for it, dfChunk in enumerate(dfChunks, start=1):
                        logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                        for columnName in dfChunk.columns:
                            logger.info(f"""  ..    Working on column "{columnName}".""")
                            if columnName in COLUMNS_TO_DE_IDENTIFY:
                                logger.info("""  ..  ..  Column must be de-identified. Collecting values.""")
                                valuesSet = sorted(list(set(dfChunk[columnName].dropna().values)))
                                if columnName in VARIABLE_ALIASES.keys():
                                    mapLookUpName = VARIABLE_ALIASES[columnName]
                                else:
                                    mapLookUpName = columnName
                                columnSetFpath = columnSetsVarsDi[mapLookUpName]["fpath"]
                                columnSetFileMode = columnSetsVarsDi[mapLookUpName]["fileMode"]
                                with open(columnSetFpath, columnSetFileMode) as file:
                                    for value in valuesSet:
                                        file.write(str(value))
                                        file.write("\n")
                                columnSetsVarsDi[mapLookUpName]["fileMode"] = "a"
                                logger.info(f"""  ..  ..  Values saved to "{columnSetFpath.absolute().relative_to(rootDirectory)}" in the project directory.""")
                else:
                    logger.info("""    This file does not need to be processed.""")

    # Return path to sets fo ID values
    # TODO If this is implemented as a function, instead of a stand-alone script, return `runOutputDir` to define `setsPathDir` in the "makeMap" scripts.
    logger.info(f"""Finished collecting the set of ID values to de-identify. The set files are located in "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")

    return runOutputDir



"""
De-identify files

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
# TODO Assign portion name to each path (per OS) so that portion files are stored in their respective folders, this prevents file from being overwritten in the unlikely, but possible, case files from different portions have the same name.
"""

# Third-party packages
import pandas as pd
# Local packages
from drapi.constants.constants import DATA_TYPES
from drapi.drapi import getTimestamp, makeDirPath, fileName2variableName, map2di, makeMap


def deIdentify(MAPS_DIR_PATH,
               listOfPortionDirs,
               LIST_OF_PORTION_CONDITIONS,
               IRB_NUMBER,
               COLUMNS_TO_DE_IDENTIFY,
               VARIABLE_ALIASES,
               VARIABLE_SUFFIXES,
               CHUNK_SIZE,
               pipelineOutputDir,
               logger,
               ROOT_DIRECTORY,
               rootDirectory):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Load de-identification maps for each variable that needs to be de-identified
    logger.info("""Loading de-identification maps for each variable that needs to be de-identified.""")
    mapsDi = {}
    mapsColumnNames = {}
    variablesCollected = [fileName2variableName(fname) for fname in MAPS_DIR_PATH.iterdir()]
    for varName in variablesCollected:
        if varName in VARIABLE_ALIASES.keys():
            map_ = makeMap(IDset=set(), IDName=varName, startFrom=1, irbNumber=IRB_NUMBER, suffix=VARIABLE_SUFFIXES[varName]["deIdIDSuffix"], columnSuffix=varName, deIdentifiedIDColumnHeaderFormatStyle="lemur")
            mapsColumnNames[varName] = map_.columns[-1]
        else:
            varPath = MAPS_DIR_PATH.joinpath(f"{varName} map.csv")
            map_ = pd.read_csv(varPath)
            mapDi = map2di(map_)
            mapsDi[varName] = mapDi
            mapsColumnNames[varName] = map_.columns[-1]

    # De-identify columns
    logger.info("""De-identifying files.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                logger.info("""  ..  Reading file to count the number of chunks.""")
                numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                logger.info(f"""  ..  This file has {numChunks} chunks.""")
                dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logger.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in COLUMNS_TO_DE_IDENTIFY:  # Keep this reference to `COLUMNS_TO_DE_IDENTIFY` as a way to make sure that all variables were collected during `getIDValues` and the `makeMap` scripts.
                            variableDataType = DATA_TYPES[columnName]
                            logger.info(f"""  ..  ..  Column must be de-identified. De-identifying values. Values are being treated as the following data type: "{variableDataType}".""")
                            if columnName in VARIABLE_ALIASES.keys():
                                mapsDiLookUpName = VARIABLE_ALIASES[columnName]
                            else:
                                mapsDiLookUpName = columnName
                            # Look up values in de-identification maps according to data type. NOTE We are relying on pandas assigning the correct data type to the look-up values in the de-identification map.
                            if variableDataType.lower() == "numeric":
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDNum: mapsDi[mapsDiLookUpName][IDNum] if not pd.isna(IDNum) else IDNum)
                            elif variableDataType.lower() == "string":
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDvalue: mapsDi[mapsDiLookUpName][str(IDvalue)] if not pd.isna(IDvalue) else IDvalue)
                            else:
                                msg = "The table column is expected to have a data type associated with it."
                                logger.error(msg)
                                raise ValueError(msg)
                            dfChunk = dfChunk.rename(columns={columnName: mapsColumnNames[columnName]})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""    This file does not need to be processed.""")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")



"""
Makes de-identification maps, building on existing maps.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
# NOTE Expects integer IDs, so no string IDs like Epic Patient IDs.
"""

import json
# Third-party packages
import pandas as pd
# Local packages
from drapi.drapi import makeDirPath, getTimestamp, makeMap, makeSetComplement, ditchFloat, handleDatetimeForJson


def makeMapsFromOthers(SETS_PATH,
                       OLD_MAPS_DIR_PATH,
                       IRB_NUMBER,
                       VARIABLE_SUFFIXES,
                       DATA_TYPES,
                       ROOT_DIRECTORY,
                       rootDirectory,
                       pipelineOutputDir,
                       logger):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    runIntermediateDataDir = pipelineOutputDir.joinpath(functionName, getTimestamp(), "temp")
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Get set of values
    # NOTE The code that used to be in this section was moved to "getIDValues.py"
    getIDValuesOutput = SETS_PATH
    logger.info(f"""Using the set of new values in the directory "{getIDValuesOutput.absolute().relative_to(rootDirectory)}".""")

    # Assertions
    collectedVariables = sorted([fname.stem for fname in getIDValuesOutput.iterdir()])
    for variableName in collectedVariables:
        assert variableName in DATA_TYPES.keys(), f"""The variable "{variableName}" was not in the `DATA_TYPES` dictionary."""
        assert variableName in VARIABLE_SUFFIXES.keys(), f"""The variable "{variableName}" was not in the `VARIABLE_SUFFIXES` dictionary."""

    # Concatenate all old maps
    oldMaps = {}
    logger.info("""Concatenating all similar pre-existing maps.""")
    for variableName in collectedVariables:
        logger.info(f"""  Working on variable "{variableName}".""")
        if variableName in OLD_MAPS_DIR_PATH.keys():
            logger.info("""    Variable has pre-existing map(s).""")
            listOfMapPaths = OLD_MAPS_DIR_PATH[variableName]
            dfConcat = pd.DataFrame()
            for mapPath in listOfMapPaths:
                logger.info(f"""  ..  Reading pre-existing map from "{mapPath}".""")
                df = pd.DataFrame(pd.read_csv(mapPath))
                dfConcat = pd.concat([dfConcat, df])
            oldMaps[variableName] = dfConcat
        elif variableName not in OLD_MAPS_DIR_PATH.keys():
            logger.info("""    Variable has no pre-existing map.""")
            oldMaps[variableName] = pd.DataFrame()

    # Get the set difference between all old maps and the set of un-mapped values
    valuesToMap = {}
    setsToMapDataDir = runIntermediateDataDir.joinpath("valuesToMap")
    makeDirPath(setsToMapDataDir)
    logger.info("""Getting the set difference between all old maps and the set of un-mapped values.""")
    for variableName in collectedVariables:
        logger.info(f"""  Working on variable "{variableName}".""")
        variableDataType = DATA_TYPES[variableName]

        # Get old set of IDs
        logger.info("""    Getting the old set of IDs.""")
        oldMap = oldMaps[variableName]
        if len(oldMap) > 0:
            oldIDSet = set(oldMap[variableName].values)
            oldIDSet = set([ditchFloat(el) for el in oldIDSet])  # NOTE: Hack. Convert values to type int or string
        elif len(oldMap) == 0:
            oldIDSet = set()
        logger.info(f"""    The size of this set is {len(oldIDSet):,}.""")

        # Get new set of IDs
        newSetPath = getIDValuesOutput.joinpath(f"{variableName}.txt")
        logger.info(f"""    Getting the new set of IDs from "{newSetPath.absolute().relative_to(rootDirectory)}".""")
        newIDSet = set()
        with open(newSetPath, "r") as file:
            text = file.read()
            lines = text.split("\n")[:-1]
        for line in lines:
            newIDSet.add(line)
        if variableDataType.lower() == "numeric":
            newIDSet = set([ditchFloat(el) for el in newIDSet])  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            pass
        else:
            msg = "The table column is expected to have a data type associated with it."
            logger.error(msg)
            raise ValueError(msg)
        logger.info(f"""    The size of this set is {len(newIDSet):,}.""")

        # Set difference
        IDSetDiff = newIDSet.difference(oldIDSet)
        valuesToMap[variableName] = IDSetDiff

        # Save new subset to `setsToMapDataDir`
        fpath = setsToMapDataDir.joinpath(f"{variableName}.JSON")
        with open(fpath, "w") as file:
            if variableDataType.lower() == "numeric":
                li = [ditchFloat(IDNumber) for IDNumber in IDSetDiff]  # NOTE: Hack. Convert values to type int or string
            elif variableDataType.lower() == "string":
                li = list(IDSetDiff)
            else:
                msg = "The table column is expected to have a data type associated with it."
                logger.error(msg)
                raise Exception(msg)
            file.write(json.dumps(li, default=handleDatetimeForJson))
        if len(IDSetDiff) == 0:
            series = pd.Series(dtype=int)
        else:
            if variableDataType.lower() == "numeric":
                series = pd.Series(sorted(list(IDSetDiff)))
            elif variableDataType.lower() == "string":
                series = pd.Series(sorted([str(el) for el in IDSetDiff]))  # NOTE TODO I don't know why this variable exists, if it was ever used, or what it's for.
            else:
                msg = "The table column is expected to have a data type associated with it."
                logger.error(msg)
                raise Exception(msg)

    # Get numbers for new map
    logger.info("""Getting numbers for new map.""")
    newNumbersDict = {}
    for variableName in collectedVariables:
        oldMap = oldMaps[variableName]
        if len(oldMap) > 0:
            oldNumbersSet = set(oldMap["deid_num"].values)
        elif len(oldMap) == 0:
            oldNumbersSet = set()
        # Get quantity of numbers needed for map
        quantityOfNumbersUnmapped = len(valuesToMap[variableName])
        # Get new numbers
        lenOlderNumbersSet = len(oldNumbersSet)
        if lenOlderNumbersSet == 0:
            newNumbers = list(range(1, quantityOfNumbersUnmapped + 1))
        else:
            newNumbersSet = makeSetComplement(oldNumbersSet, quantityOfNumbersUnmapped)
            newNumbers = sorted(list(newNumbersSet))
        newNumbersDict[variableName] = newNumbers

    # Map un-mapped values
    logger.info("""Mapping un-mapped values.""")
    for file in setsToMapDataDir.iterdir():
        variableName = file.stem
        variableDataType = DATA_TYPES[variableName]
        logger.info(f"""  Working on un-mapped values for variable "{variableName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Map contents
        values = valuesToMap[variableName]
        if variableDataType.lower() == "numeric":
            values = set(int(float(value)) for value in values)  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            pass
        else:
            msg = "The table column is expected to have a data type associated with it."
            logger.error(msg)
            raise Exception(msg)
        numbers = sorted(list(newNumbersDict[variableName]))
        deIdIDSuffix = VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"]
        map_ = makeMap(IDset=values, IDName=variableName, startFrom=numbers, irbNumber=IRB_NUMBER, suffix=deIdIDSuffix, columnSuffix=variableName, deIdentificationMapStyle="lemur")
        # Save map
        mapPath = runOutputDir.joinpath(f"{variableName} map.csv")
        map_.to_csv(mapPath, index=False)
        logger.info(f"""    De-identification map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")

    return runOutputDir



"""
Consent-to-Share helper functions. Consent-to-Share is also known as "C2S".
"""

import logging
import os
from pathlib import Path
from typing import List
from typing_extensions import Literal
# Third-party libraries
import pandas as pd
# Local libraries
from drapi.code.drapi.drapi import (replace_sql_query,
                                    successiveParents)

# Arguments
MODULE_ROOT_DIRECTORY_PATH, _ = successiveParents(Path(__file__), 4)

# Arguments: SQL connection settings
SERVER = "EDW.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"


def checkStatus(statusType=Literal["C2S", "death"],
                location=Literal["gnv", "jax"],
                listOfMRNs=List[str]) -> pd.DataFrame:
    """
    This functions performs the Consent-to-Share ("C2S") check before release, to ensure deceased or opted-out patients aren't contacted.
    """

    # Determine query type
    if statusType == "C2S":
        queryFilePath = MODULE_ROOT_DIRECTORY_PATH.joinpath("sql/Consent2Share.sql")
    elif statusType == "death":
        queryFilePath = MODULE_ROOT_DIRECTORY_PATH.joinpath("sql/LADMF.sql")

    # Define values for query: LOCATION_NAME, LOCATION_TYPE
    location_lower_case = location.lower()
    if location_lower_case == "gnv":
        locationNameForQuery = "UF"
        locationValueForQuery = "101"
    elif location_lower_case == "jax":
        locationNameForQuery = "Jax"
        locationValueForQuery = "110"

    # Define value for query: LIST_OF_MRNS
    MRNValuesForQuery = ",".join(f"'{MRNNumber}'" for MRNNumber in listOfMRNs)

    # Load query template
    with open(queryFilePath, "r") as file:
        query0 = file.read()

    # Prepare query
    query = replace_sql_query(query=query0, old="{<PYTHON_PLACEHOLDER : LOCATION_NAME>}", new=locationNameForQuery)
    query = replace_sql_query(query=query, old="{<PYTHON_PLACEHOLDER : LOCATION_TYPE>}", new=locationValueForQuery)
    query = replace_sql_query(query=query, old="{<PYTHON_PLACEHOLDER : LIST_OF_MRNS>}", new=MRNValuesForQuery)

    # Run query
    logging.debug(query)
    queryResult = pd.read_sql(sql=query, con=conStr)

    return queryResult


def doCheck(result, statusType):
    """
    Analyzes table values to deduce if the entire table passes the check. Any patients that fail the check are returned in a table.
    """
    result = pd.DataFrame(result)
    if statusType == "C2S":
        s1 = pd.Series(result["CONSENT_SHARE_IND"])
        r1 = (s1 == "Y")
        r1sum = r1.sum()
        s2 = pd.Series(result["CONSENT_SHARE_DATE"])
        r2 = s2.notna()
        r2sum = r2.sum()
        length = len(result)
        if (r1sum == length) and (r2sum == length):
            checkResultPass = True
            failedRows = None
        else:
            checkResultPass = False
            # Print failed rows
            mask = ~r1 | ~r2
            failedRows = result[mask]

    elif statusType == "death":
        s1 = pd.Series(result["PATNT_SSN_DTH_IND"])
        r1 = (s1 == "N")
        r1sum = r1.sum()
        s2 = pd.Series(result["PATNT_SSN_DTH_DATE"])
        r2 = s2.isna()
        r2sum = r2.sum()
        s3 = pd.Series(result["PATNT_DTH_IND"])
        r3 = (s3 == "N")
        r3sum = r3.sum()
        s4 = pd.Series(result["PATNT_DTH_DATE"])
        r4 = s4.isna()
        r4sum = r4.sum()
        length = len(result)
        if (r1sum == length) and (r2sum == length) and (r3sum == length) and (r4sum == length):
            checkResultPass = True
            failedRows = None
        else:
            checkResultPass = False
            # Print failed rows
            mask = ~r1 | ~r2 | ~r3 | ~r4
            failedRows = result[mask]

    return checkResultPass, failedRows



"""
Utility functions commonly used in IDR data request projects.
"""

import datetime as dt
import logging
import os
import re
import sys
from array import array
from collections.abc import Collection
from datetime import datetime
from datetime import date
from datetime import time
from dateutil.parser import parse
from itertools import islice
from logging import Logger
from pathlib import Path
from time import sleep
from typing import Callable, List, Tuple, Union
from typing_extensions import Literal
# Third-party packages
import numpy as np
import pandas as pd
import sqlalchemy as sa
import sqlite3
from pandas.io.parsers.readers import TextFileReader
# Local packages
pass

logger = logging.getLogger(__name__)

# SQL Server settings
SERVER = "DWSRSRCH01.shands.ufl.edu"  # AKA `HOST`
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = fr"{USERDOMAIN}\{USERNAME}"
PWD = os.environ["HFA_UFADPWD"]

# SQLAlchemy connections
conStr = f"mssql+pymssql://{UID}:{PWD}@{SERVER}/{DATABASE}"  # Create connection string
engine = sa.create_engine(conStr)  # Make connection/engine


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    h/t to https://stackoverflow.com/a/39215961/5478086
    """

    def __init__(self, logger, level):
        self.logger = logger
        self.level = level
        self.linebuf = ''

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())

    def flush(self):
        pass


class LoggerWriter:
    """
    h/t to https://stackoverflow.com/a/31688396/5478086
    """

    def __init__(self, level):
        self.level = level

    def write(self, message):
        if message != '\n':
            self.level(message)

    def flush(self):
        self.level(sys.stderr)


def makeDirPath(directory_path: str) -> None:
    """
    Check if all directories exists in a path. If not, create them
    """
    path_obj = Path(directory_path)
    paths = list(path_obj.parents)[::-1] + [path_obj]
    for dir in paths:
        if not os.path.exists(dir):
            os.mkdir(dir)


def getTimestamp():
    return dt.datetime.now().strftime("%Y-%m-%d %H-%M-%S")


def getPercentDifference(x, y):
    """
    Calculates a percentage if the denominator is not 0, else it returns `"N/A"`.
    """
    if y != 0:
        return f"{x / y: 0.2%}"
    else:
        return "N/A"


def flatExtend(iterable: Collection) -> list:
    """
    Creates a list from an iterable.
    """
    outputList = []
    for el in iterable:
        outputList.extend(el)
    return outputList


def readDataFile(fname: Path, *args, **kwargs) -> TextFileReader:
    """
    Opens a file based on the file extension of its file name `fname` using a Pandas built-in to return a DataFrame object. Below is the list of Pandas built-ins, whose odcumentation you can search to find out what keyword arguments (`kwargs`) to use.
      - `pd.read_csv`
      - `pd.read_json`
      - `pd.read_xml`
      - `pd.read_excel`
      - `pd.read_hdf`
      - `pd.read_sql`
    """
    suffix = fname.suffix.lower()
    if suffix.endswith(('.csv',)):
        TextFileReaderObject = pd.read_csv(fname, *args, **kwargs)
    elif suffix.endswith(('.tsv',)):
        kwargs["delimiter"] = "\t"
        TextFileReaderObject = pd.read_csv(fname, *args, **kwargs)
    elif suffix.endswith(('.json',)):
        TextFileReaderObject = pd.read_json(fname, *args, **kwargs)
    elif suffix.endswith(('.xml',)):
        TextFileReaderObject = pd.read_xml(fname, *args, **kwargs)
    elif suffix.endswith(('.xls', 'xlsx',)):
        TextFileReaderObject = pd.read_excel(fname, *args, **kwargs)
    elif suffix.endswith(('.hdf',)):
        TextFileReaderObject = pd.read_hdf(fname, *args, **kwargs)
    elif suffix.endswith(('.sql',)):
        TextFileReaderObject = pd.read_sql(fname, *args, **kwargs)
    else:
        raise ValueError(f"""Unsupported filetype for file "{fname}".""")
    return TextFileReaderObject


def successiveParents(pathObj: Path, numLevels: int) -> Tuple[Path, int]:
    """
    Successively get the parents of the Path object submitted.
    """
    while numLevels > 0:
        pathObj = pathObj.parent
        numLevels -= 1
    return pathObj, numLevels


def fileContainsColumn(pathObj: Path, listOfColumns: list) -> bool:
    """
    Determines if the file defined by `pathObj` has any of the columns contained in `listOfColumns`.
    """
    condition = False
    df = pd.read_csv(pathObj, nrows=10)
    columns = df.columns
    for columnName in listOfColumns:
        if columnName in columns:
            condition = True
            break
        else:
            continue
    return condition


def getCommonDirectoryParent(primaryPath: Path, secondaryPath: Path) -> Path:
    """
    """
    commonPrefix = os.path.commonprefix([primaryPath, secondaryPath])
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        if commonPrefix == "/":
            pathResult = primaryPath
        else:
            pathResult = primaryPath.absolute().relative_to(secondaryPath)
    elif operatingSystem == "win32":
        if commonPrefix == "":
            pathResult = primaryPath
        else:
            pathResult = primaryPath.absolute().relative_to(secondaryPath)
    return pathResult


def choosePathToLog(path: Path, rootPath: Path) -> Path:
    """
    Decides if a path is a subpath of `rootPath`. If it is, display it reltaive to `rootPath`. If it is not, display it as an absolute path.
    """
    commonPath = os.path.commonpath([path.absolute(), rootPath.absolute()])

    lenCommonPath = len(commonPath)
    lenRootPath = len(str(rootPath.absolute()))
    if lenCommonPath < lenRootPath:
        pathToDisplay = path
    elif lenCommonPath >= lenRootPath:
        pathToDisplay = path.absolute().relative_to(rootPath)
    else:
        raise Exception("An unexpected error occurred.")

    return pathToDisplay


def getMapType(df: pd.DataFrame) -> Literal["1:1", "1:m", "m:1", "m:m"]:
    """
    h/t to https://stackoverflow.com/questions/59091196
    """
    df = df.drop_duplicates()
    dfShape0 = df.shape
    assert dfShape0[1] == 2, "Maps are supposed to have only two columns."
    col1Name, col2Name = df.columns

    first_max = df.groupby(col2Name).count().max().loc[col1Name]
    second_max = df.groupby(col1Name).count().max().loc[col2Name]
    if first_max == 1:
        if second_max == 1:
            return "1:1"
        else:
            return "1:m"
    else:
        if second_max == 1:
            return "m:1"
        else:
            return "m:m"


def getFilesToRelease(filesToRelease: List[Path], fileCriteria: List[Callable]):
    """
    Applies the functions in the iterable `fileCriteria` to each Path object in `filesToRelease`.
    """
    filesToRelease = []
    for file in filesToRelease:
        criteria = []
        for criteriaFunc in fileCriteria:
            criteria.append(criteriaFunc(file))
        criteriaMet = all(criteria)
        if criteriaMet:
            filesToRelease.append(file)
        else:
            pass


def getLastIDNum(df: pd.DataFrame, columnName="deid_num"):
    """
    Gets the last ID number in a de-identification map.
    """
    numbers = df[columnName].astype(int)
    return max(numbers)


def isNumber(string):
    """
    Checks if the string represents a number

    For a discussion see https://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-represents-a-number-float-or-int
    """
    try:
        int(float(string))
        return True
    except ValueError as error:
        msg = error.args[0]
        if "invalid literal for int() with base 10:" in msg:  # This must be an old Python exception message
            return False
        elif "could not convert string to float:" in msg:  # I think this is the new exception message
            return False
        else:
            raise Exception(error)


def getSerialNumber(string: str):
    """
    Extracts the serial number of a file name using regular expressions.
    """
    reObj = re.search("[0-9]+", string)
    if reObj:
        result = reObj.group()
    else:
        result = "0"
    return int(result)


def fileName2variableName(pathObj):
    """
    """
    pattern = r"^(?P<variableName>.+) map"
    string = pathObj.stem
    obj = re.match(pattern, string)
    if obj:
        groupDict = obj.groupdict()
        variableName = groupDict["variableName"]
    else:
        raise Exception("This file path objected was of an unexpected format.")
    return variableName


def mapGroupCriteria4unknownValue(value):
    """
    Checks if `value` is a number. If so, it checks if it's less than 0. If it's not a number, return `True`.

    This is used in making de-identification maps, where negative ID values are usually used to represent unknown or null values.
    """
    if isNumber(value):
        return float(value) < 0
    else:
        return False


def makeMap(IDset: set,
            IDName: str,
            startFrom: Union[int, list],
            irbNumber: str,
            suffix: str,
            columnSuffix: str,
            logger: Logger,
            groups: dict = {0: {"criteria": [mapGroupCriteria4unknownValue],
                                "deidNum": 0}},
            deIdentificationMapStyle: Literal["classic", "lemur"] = "lemur") -> pd.DataFrame:
    """
    Makes an IDR de-identification map.

    INPUT
        `IDset`, a set of IDs
        `IDName`, the name of the ID
        `startFrom`, the integer number to start from
        `groups`, ID values to group or map in a many-to-one fashion. E.g., invalid IDs (negative numbers) are usually all mapped to the same de-identified number, like "0".
        `deIdentificationMapStyle`, a string, one of {"classic, "lemur"}. The formats are as follow:
            | Format Style  | de-Identififed ID Column Header   |
            | ------------- | -------------------------------   |
            | "clasic"      | `f"deid_{columnSuffix}_id"`       |
            | "lemur"       | `f"de-Identified {IDName}"`       |

    OUTPUT
        `map_`, a Pandas DataFrame with the following format:
        | `IDName` | deid_num   | de-Identififed ID Column Header |
        | -------- | --------   | ------------------------ |
        | IDset[0] | numbers[0] | ... |
        | ...      | ...        | ... |
    """
    # QA: `irbNumber` must not be NoneType
    if isinstance(irbNumber, type(None)):
        raise Exception("""`irbNumber` cannot be of type "None".""")

    # Assign header formats: de-Identififed ID Column Header
    if deIdentificationMapStyle == "classic":
        deIdentificationSerialNumberHeader = "deid_num"
    elif deIdentificationMapStyle == "lemur":
        deIdentificationSerialNumberHeader = "De-identification Serial Number"

    # Assign header formats: de-Identififed ID Column Header
    if deIdentificationMapStyle == "classic":
        deIdentifiedIDColumnHeader = f"deid_{columnSuffix}_id"
    elif deIdentificationMapStyle == "lemur":
        deIdentifiedIDColumnHeader = f"De-identified {IDName}"

    if len(IDset) == 0:
        return pd.DataFrame(columns=[IDName,
                                     deIdentificationSerialNumberHeader,
                                     deIdentifiedIDColumnHeader])
    else:
        pass
    if isinstance(startFrom, int):
        startFrom = startFrom
        numbers = list(range(startFrom, startFrom + len(IDset)))
    elif isinstance(startFrom, list):
        numbers = startFrom[:]
        startFrom = numbers[0]
    numbers.extend([None])
    logger.info("    Sorting `IDset`.")
    IDli = sortIntegersAndStrings(list(IDset))
    logger.info("    Sorting `IDset` - done.")
    logger.info("    Creating `mapDi`.")
    mapDi = {IDNum: {} for IDNum in IDli}
    logger.info("    Creating `mapDi` - done.")
    lenIDli = len(IDli)
    if lenIDli > 100000:
        itChunk = 1000
    else:
        itChunk = max(round(lenIDli / 50), 1)
    for it, IDNum in enumerate(IDli, start=1):
        fromGroup = False
        for group, groupAttributes in groups.items():
            criteriaList = groupAttributes["criteria"]
            criteria = [criterion(IDNum) for criterion in criteriaList]
            if all(criteria):
                deid_num = groupAttributes["deidNum"]
                fromGroup = True
                break
        if fromGroup:
            pass
        else:
            deid_num = numbers.pop(0)
        deid_id = f"{irbNumber}_{suffix}_{deid_num}"
        mapDi[IDNum] = {IDName: IDNum,
                        deIdentificationSerialNumberHeader: deid_num,
                        deIdentifiedIDColumnHeader: deid_id}
        if it % itChunk == 0:
            logger.info(f"  ..  Working on item {it:,} of {len(IDli):,} ({it/lenIDli:.2%} done).")
    newMap = pd.DataFrame.from_dict(mapDi, orient="index")
    newMap.index = range(1, len(newMap) + 1)
    return newMap


def makeSetComplement(set1, cardinalityOfNewSet):
    """
    Creates a complement to a set, based on the specified size of the complement. Elements of the first set are assumed to be integers from 1 and above, and cardinalities are non-zero positive integers.

    TODO: Implement empty set handling
    """
    set1Min = 1
    if len(set1) == 0:
        raise ValueError("Set should not be empty.")
    else:
        set1Max = max(set1)
    s1Contiguous = set(range(set1Min, set1Max + 1))
    setDifference = s1Contiguous.difference(set1)
    cardinalityOfSetDifference = len(setDifference)
    cardinalityOfContiguousSubset = cardinalityOfNewSet - cardinalityOfSetDifference
    if cardinalityOfContiguousSubset > 0:
        setDifferenceSubset = setDifference
        contiguousSubset = set(range(set1Max + 1, set1Max + cardinalityOfContiguousSubset + 1))
    elif cardinalityOfContiguousSubset == 0:
        setDifferenceSubset = setDifference
        contiguousSubset = set()
    elif cardinalityOfContiguousSubset < 0:
        setDifferenceSubset = set(list(setDifference)[:cardinalityOfNewSet])
        contiguousSubset = set()
    newSet = setDifferenceSubset.union(contiguousSubset)
    return newSet


def loglevel2int(loglevel: Union[int, str]) -> int:
    """
    An agnostic converter that takes int or str and returns int. If input is int, output is the same as input"""
    dummy = logging.getLogger("dummy")
    dummy.setLevel(loglevel)
    loglevel = dummy.level
    return loglevel


def replace_sql_query(query: str, old: str, new: str, logger: logging.Logger) -> str:
    """
    Replaces text in a SQL query only if it's not commented out. I.e., this function applies string.replace() only if the string doesn't begin with "--".
    TODO Don't replace text after "--".
    """
    pattern = r"^\w*--"
    li = query.split("\n")
    result = []
    logger.debug("Starting")
    for line in li:
        logger.debug(f"""  Working on "{line}".""")
        obj = re.search(pattern, line)
        if obj:
            logger.debug("    Passing")
            nline = line
        else:
            nline = line.replace(old, new)
            logger.debug(f"""    Replacing text in "{line}" --> "{nline}".""")
        logger.debug("  Appending...")
        result.append(nline)
    logger.debug("Finished")
    return "\n".join(result)


def sqlite2df(tableContents: list, tableName: str, cursor: sqlite3.Cursor) -> pd.DataFrame:
    """
    `tableContents` must be a "*" query where all the columns are returned.
    """
    # Get metadata on table
    query = f"PRAGMA table_info({tableName});"
    cursor.execute(query)
    query_results = cursor.fetchall()
    column_header = [tu[1] for tu in query_results]

    tableContents_asDict = {}
    for it, table_info in enumerate(tableContents):
        di = {}
        for key, value in zip(column_header, table_info):
            di[key] = value
        tableContents_asDict[it] = di
    df = pd.DataFrame.from_dict(tableContents_asDict, orient="index")
    df.index = range(1, len(tableContents) + 1)
    return df


def patient_key_from_person_id(person_id: int, map_: dict = {}) -> Tuple[int, Literal[0, 1]]:
    """
    Assumes "map_" is a dictionary with person IDs as integers that map patient keys as integers.
    """
    if person_id in map_.keys():
        patient_key = map_[person_id]
        source = 0
    else:
        query = f"""use DWS_OMOP_PROD
        SELECT
            xref.PERSON_MAPPING.person_id as person_id,
            xref.PERSON_MAPPING.patient_key as patient_key
        FROM
            xref.PERSON_MAPPING
        WHERE
            xref.PERSON_MAPPING.person_id IN ({person_id})"""
        results = pd.read_sql(query, con=conStr)
        person_id = results["person_id"][0]
        patient_key = results["patient_key"][0]
        source = 1
    return patient_key, source


def personIDs2patientKeys(personIDList: List[int]) -> pd.DataFrame:
    """

    """
    list2str = ",".join([str(num) for num in personIDList])
    query = f"""use DWS_OMOP_PROD
    SELECT
        xref.PERSON_MAPPING.person_id as person_id,
        xref.PERSON_MAPPING.patient_key as patient_key
    FROM
        xref.PERSON_MAPPING
    WHERE
        xref.PERSON_MAPPING.person_id IN ({list2str})"""
    results = pd.read_sql(query, con=conStr)
    return results


def epicID2patientKey(epicIDlist: List[str]) -> pd.DataFrame:
    """
    Executes a SQL query to map EPIC Patient IDs to Patient Keys.
    """
    IDsAsString = ",".join([f"'{el}'" for el in epicIDlist])
    query = f"""USE DWS_PROD
    SELECT
        dbo.ALL_PATIENTS.PATNT_KEY AS 'Patient Key',
        dbo.ALL_PATIENTS.PATNT_ID AS 'EPIC Patient ID'
    FROM
        dbo.ALL_PATIENTS
    WHERE
        (
        dbo.ALL_PATIENTS.PATNT_ID  IN  ( {IDsAsString}  )
        AND
        ( dbo.ALL_PATIENTS.TEST_IND='N'  )
        )"""
    results = pd.read_sql(query, con=conStr)
    return results


def map2di(map_: pd.DataFrame, fromColumnIndex: int = 0, toColumnIndex: int = 2) -> dict:
    """
    `fromColumnIndex` is the variable to assign as the key to the dictionary.
    `toColumnIndex` is the variable to assign as the value pair in the dictionary.
    Assumes the following structure to "map_":
    column 1      | column 2       | column 3      | ... | column k
        x_(1,1)   |     x_(1,2)    |    x_(1,3)    | ... |   x_(1,k)
        .         |     .          |     .         | ... |   .
        .         |     .          |     .         | ... |   .
        .         |     .          |     .         | ... |   .
        x_(n,1)   |     x_(n,2)    |    x_(n,3)    | ... |   x_(n,k)
    Where the columns are from 1, 2, 3, ..., i, ..., k-1, k, and the rows are from 1, 2, 3, ..., j, ..., n-1, n.
    """
    di = {}
    errors = []
    for _, row in map_.iterrows():
        oldValue = row.values[fromColumnIndex]
        newValue = row.values[toColumnIndex]
        if oldValue in di.keys() and di[oldValue] != newValue:
            errors.append({oldValue: newValue})
        di[oldValue] = newValue
    if len(errors) > 0:
        raise Exception(f"This map is not one to one. At least one ID is duplicated with multiple de-identified IDs: {errors}")
    else:
        return di


def float2str(value: float, navalue="") -> str:
    """
    Converts values to string-type. If a value is NaN it is replaced with a value that can be converted to a string, default empty string, "".
    """
    assert isinstance(navalue, str), """"navalue" is not a string type."""
    if pd.isna(value):
        newValue = str(navalue)
    else:
        newValue = str(int(value))
    return newValue


def numericOrString2integerOrString(value) -> Union[str, int]:
    """
    Determines if a value is numeric or string. If numeric, convert to an integer, else return as string.

    Used by `makeMapsFromOthers.py`
    """
    if isNumber(value):
        return int(float(value))
    else:
        return str(value)


def sortIntegersAndStrings(li: list) -> list:
    """
    Sorts a list of integers and strings by sorted them separately and then concatenating the results.
    """
    series = pd.Series(li)
    maskIntegers = series.apply(lambda el: isinstance(el, int))
    maskStrings = series.apply(lambda el: isinstance(el, str))
    li2 = sorted(series[maskIntegers].to_list()) + sorted(series[maskStrings].to_list())
    return li2


def str2int(value, navalue=-1):
    """
    Converts values to integer-type. If a value is NaN it is replaced with a value that can be converted to an integer, default is "-1".
    """
    assert isinstance(navalue, int), """"navalue" is not an integer type."""
    if pd.isna(value):
        newValue = int(navalue)
    else:
        newValue = int(float(value))
    return newValue


def str2intOr0(string):
    """
    Converts a string to its numeric value, or 0, if it has no numeric value.
    """
    if string.isnumeric():
        return int(string)
    else:
        return 0


def str2bool(value, navalue=""):
    """
    Converts values to boolean-type. If a value is missing it is replaced with an empty string.
    """
    if pd.isna(value):
        newValue = int(navalue)
    else:
        newValue = bool(value)
    return newValue


def deconstructStudyID(string):
    """
    Deconstructs a de-identified study ID into its component parts, assuming the following format: STUDY-TYPESTUDY-NUMBER_SERIAL-NUMBER
    Where "STUDY_TYPE" is the study type, e.g., "CED", "IRB", or "WIRB"
          "STUDY-NUMBER" is the study serial number, usually a 9-digit number where the first four numbers is a four-digit year.
          "SERIAL-NUMBER" is the de-identified study-specific serial number of a patient or other subject.
    For example: "IRB202300123_222"
    """
    pattern = r"(?P<studyType>\w+)(?P<studyNumber>\d{9})_(?P<PHI_type>\w+)_(?P<deIdentificationNumber>\d+)"
    regexObj = re.search(pattern, string)
    if regexObj:
        di = regexObj.groupdict()
        return di
    else:
        raise Exception("String is of an unexpected format. See function docstring for expected format.")


def studyID2tuple(string):
    """
    Extracts the numeric portion of a string.

    Assumes the following format: STUDY-TYPESTUDY-NUMBER_SERIAL-NUMBER
    Where "STUDY_TYPE" is the study type, e.g., "CED", "IRB", or "WIRB"
          "STUDY-NUMBER" is the study serial number, usually a 9-digit number where the first four numbers is a four-digit year.
          "SERIAL-NUMBER" is the de-identified study-specific serial number of a patient or other subject.
    For example: "IRB202300123_222"

    This is intended for use when sorting columns by the numeric value of the string, and not the text value.
    """
    di = deconstructStudyID(string)
    studyType = di["studyType"]
    studyNumber = int(di["studyNumber"])
    PHI_type = di["PHI_type"]
    deIdentificationNumber = int(di["deIdentificationNumber"])
    tu = studyType, studyNumber, PHI_type, deIdentificationNumber
    return tu


def isValidPatientID(value):
    """
    Checks if the value is a valid UFHealth patient ID. Invalid patient IDs are defined as negative integers
    """
    isNumber_bool = isNumber(value)
    if isNumber_bool:
        number = int(value)
        if number < 0:
            result = False
        elif number >= 0:  # Assumes `0` is a valid patient ID
            result = True
    elif not isNumber_bool:
        result = True  # Assumes all non-numeric IDs are valid
    return result


def isDatetime(string):
    """
    Checks if a string contains a datetime format.
    """
    try:
        _ = parse(string, fuzzy=False)
        return True
    except ValueError:
        return False


def ditchFloat(value):
    """
    Gets rid of float types if value is string, integer, float, or datetime
    """
    if isinstance(value, str):
        value = str(value)
        if value.isnumeric():
            standardValue = int(value)
        elif isNumber(value):
            standardValue = int(float(value))
        elif isDatetime(value):
            standardValue = parse(value, fuzzy=False)
        else:
            raise ValueError(f"""Unexpected format in string value "{value}". We expected a string with a number, numeric, or datetime format.""")
    elif isinstance(value, int) or isinstance(value, np.integer) or isinstance(value, float):
        standardValue = int(value)
    else:
        raise ValueError(f"""Unexpected data type "{type(value)}" for value "{value}". Expect only string, float, or integers.""")
    return standardValue


def handleDatetimeForJson(obj):
    """
    JSON serializer for objects not serializable by default json code

    h/t to https://stackoverflow.com/a/22238613/5478086
    """
    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


def makeChunks(array_range: array, chunkSize: int) -> tuple:
    """
    Inspired by GeekForGeeks.com (https://www.geeksforgeeks.org/break-list-chunks-size-n-python/)

    Only makes chunks of one dimensional arrays (e.g., lists, Pandas Series), but not dataframes.

    Example:
    array = range(30)
    chunks = makeChunks(array, chunkSize)
    """
    array_range = iter(array_range)
    return iter(lambda: tuple(islice(array_range, chunkSize)), ())


# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# >>> tree function >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# h/t https://stackoverflow.com/a/59109706/5478086


def tree(dir_path: Path,
         level: int = -1,
         limit_to_directories: bool = False,
         length_limit: int = 1000):
    """Given a directory Path object print a visual tree structure"""
    # prefix components:
    space = '    '
    branch = '│   '
    # pointers:
    tee = '├── '
    last = '└── '

    dir_path = Path(dir_path)  # accept string coerceable to Path
    files = 0
    directories = 0

    def inner(dir_path: Path, prefix: str = '', level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        if limit_to_directories:
            contents = [d for d in dir_path.iterdir() if d.is_dir()]
        else:
            contents = list(dir_path.iterdir())
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                yield prefix + pointer + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1
    print(dir_path.name)
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        print(line)
    if next(iterator, None):
        print(f'... length_limit, {length_limit}, reached, counted:')
    print(f'\n{directories} directories' + (f', {files} files' if files else ''))


# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



"""
Get all variables/columns of tables/files in the project.
"""

import os
from collections import OrderedDict
from pathlib import Path
# Third-party packages
import pandas as pd


def getProjectColumns(listOfPortionDirs: list,
                      ROOT_DIRECTORY,
                      rootDirectory,
                      logger):
    """
    """
    functionName = __name__.split(".")[-1]
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Get columns
    columns = {}
    for portionPath in listOfPortionDirs:
        content_paths = [Path(dirObj) for dirObj in os.scandir(portionPath)]
        dirRelativePath = portionPath.absolute().relative_to(rootDirectory)
        logger.info(f"""Reading files from the directory "{dirRelativePath}". Below are its contents:""")
        for fpath in sorted(content_paths):
            logger.info(f"""  {fpath.name}""")
        for file in content_paths:
            conditions = [lambda x: x.is_file(), lambda x: x.suffix == ".csv", lambda x: x.name != ".DS_Store"]
            conditionResults = [func(file) for func in conditions]
            if all(conditionResults):
                logger.info(f"""  Reading "{file.absolute().relative_to(rootDirectory)}".""")
                df = pd.read_csv(file, dtype=str, nrows=10)
                columns[file.name] = df.columns

    # Get all columns by file
    logger.info("""Printing columns by file.""")
    allColumns = set()
    it = 0
    columnsOrdered = OrderedDict(sorted(columns.items()))
    for key, value in columnsOrdered.items():
        if it > -1:
            logger.info(key)
            logger.info("")
            for el in sorted(value):
                logger.info(f"  {el}")
                allColumns.add(el)
            logger.info("")
        it += 1

    # Get all columns by portion
    # TODO
    pass

    # Print the set of all columns
    logger.info("""Printing the set of all columns.""")
    for el in sorted(list(allColumns)):
        logger.info(f"  {el}")

    # End script
    logger.info(f"""Finished running "{functionName}".""")



"""
Compare two cohort (group) files.
"""

import logging
from typing_extensions import (Literal,
                               Union)
# Third-party packages
import pandas as pd


def mappingAnalysis(x0: Union[pd.DataFrame, pd.Series], m0: pd.DataFrame, logger: logging.Logger, mappingMethod: Literal["inner", "outer"] = "inner") -> None:
    """
    # x0: The vector of input values
    # m0: The matrix of mapping values, an n_m by 2 matrix
    # x1: The vector of output values
    """
    assert len(x0.shape) == 1 or x0.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."
    assert len(m0.shape) == 2, "The mapping table should be of two dimensions."
    assert m0.shape[1] == 2, "The mapping table should be of two columns."
    if isinstance(x0, pd.DataFrame):
        pass
    elif isinstance(x0, pd.Series):
        x0 = pd.DataFrame(x0)
    else:
        raise Exception(f"""`x0` is of an unexpected type: "{type(x0)}".""")
    fromColumn = x0.columns.to_list()[0]
    toColumn = m0.columns.to_list()[1]

    mappingMethodsList = ["inner", "outer"]
    x1 = {}
    for mappingMethod in mappingMethodsList:
        x1[mappingMethod] = x0.set_index(fromColumn).join(other=m0.set_index(fromColumn),
                                                          how=mappingMethod,
                                                          lsuffix="_L",
                                                          rsuffix="_R")

    lenx0 = len(x0)
    lenx0unique = len(x0.drop_duplicates())
    lenm0 = len(m0)
    lenm0_1 = len(m0[fromColumn])
    lenm0_2 = len(m0[toColumn])
    lenm0_1unique = len(m0[fromColumn].drop_duplicates())
    lenm0_2unique = len(m0[toColumn].drop_duplicates())
    lenx1 = {}
    lenx1unique = {}
    for mappingMethod in mappingMethodsList:
        lenx1[mappingMethod] = len(x1[mappingMethod])
        lenx1unique[mappingMethod] = len(x1[mappingMethod].drop_duplicates())

    logger.info(f"""The size of the input group is {lenx0:,}.""")
    logger.info(f"""The size of the input group (unique values) is {lenx0unique:,}.""")
    logger.info(f"""The size of the entire map is {lenm0:,}.""")
    logger.info(f"""The size of the map's left side is {lenm0_1:,}.""")
    logger.info(f"""The size of the map's right side is {lenm0_2:,}.""")
    logger.info(f"""The size of the map's left side (unique values) is {lenm0_1unique:,}.""")
    logger.info(f"""The size of the map's right side (unique values) is {lenm0_2unique:,}.""")
    for mappingMethod in mappingMethodsList:
        logger.info(f"""  Mapping values using method "{mappingMethod}".""")
        logger.info(f"""    The size of the output group is {lenx1[mappingMethod]:,}.""")
        logger.info(f"""    The size of the output group (unique values) is {lenx1unique[mappingMethod]:,}.""")


def compareGroups(group1: Union[pd.DataFrame, pd.Series], group2: Union[pd.DataFrame, pd.Series], logger: logging.Logger) -> None:
    """
    """
    assert len(group1.shape) == 1 or group1.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."
    assert len(group2.shape) == 1 or group2.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."
    if isinstance(group1, pd.DataFrame):
        group1series = group1.iloc[:, 0]
    else:
        group1series = group1
    if isinstance(group2, pd.DataFrame):
        group2series = group2.iloc[:, 0]
    else:
        group2series = group2

    c1l = len(group1series)
    c2l = len(group2series)

    c1inc2 = group1series.isin(group2series)
    c2inc1 = group2series.isin(group1series)

    c1inc2count = c1inc2.sum()
    c2inc1count = c2inc1.sum()

    c1inc2percent = c1inc2count / c1l
    c2inc1percent = c2inc1count / c2l

    intersection = set(group1series.to_list()).intersection(set(group2series.to_list()))
    intersectionl = len(intersection)

    logger.info("""The following results are of the set of groups 1 and 2. I.e., we are only counting the unique values.""")
    logger.info(f"""Group 1 size: {c1l:,}.""")
    logger.info(f"""Group 2 size: {c2l:,}.""")
    logger.info(f"""Size of intersection of groups 1 and 2: {intersectionl:,}.""")
    logger.info(f"""Percent of group 1 in group 2: {c1inc2percent:.2%}.""")
    logger.info(f"""Percent of group 2 in group 1: {c2inc1percent:.2%}.""")


def determineMapType(x0: Union[pd.DataFrame, pd.Series], x1: Union[pd.DataFrame, pd.Series]):
    """
    """
    assert len(x0.shape) == 1 or x0.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."
    assert len(x1.shape) == 1 or x1.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."

    x0duplicated = x0.duplicated().sum() > 0
    x1duplicated = x1.duplicated().sum() > 0

    if x0duplicated and x1duplicated:
        mapType = "m:m"
    elif x0duplicated:
        mapType = "m:1"
    elif x1duplicated:
        mapType = "1:m"
    elif (not x0duplicated) & (not x1duplicated):
        mapType = "1:1"
    else:
        raise Exception("Unexpected error.")

    return mapType


def _determineMapType_MapValues(x0: Union[pd.DataFrame, pd.Series],
                                m0: pd.DataFrame) -> pd.DataFrame:
    """
    """
    assert len(x0.shape) == 1 or x0.shape[1] == 1, "The input vector should be of one dimension or two dimensions with one column."
    assert len(m0.shape) == 2, "The mapping table should be of two dimensions."
    assert m0.shape[1] == 2, "The mapping table should be of two columns."
    if isinstance(x0, pd.DataFrame):
        pass
    elif isinstance(x0, pd.Series):
        x0 = pd.DataFrame(x0)
    else:
        raise Exception(f"""`x0` is of an unexpected type: "{type(x0)}".""")
    fromColumn = x0.columns.to_list()[0]
    toColumn = m0.columns.to_list()[1]
    _ = toColumn

    # Map values
    x0 = pd.DataFrame(x0)  # To allow use of `set_index`
    x1 = x0.set_index(fromColumn).join(other=m0.set_index(fromColumn),
                                       how="inner",
                                       lsuffix="_L",
                                       rsuffix="_R")
    return x1


def determineMapTypeFromMap(x0: Union[pd.DataFrame, pd.Series],
                            m0: pd.DataFrame,
                            logger: Union[logging.Logger, None] = None,
                            x1: Union[pd.DataFrame, pd.Series, None] = None) -> Literal["1:1", "m:1", "1:m"]:
    """
    """
    if x1:
        x1 = x1
    elif isinstance(x1, type(None)):
        x1 = _determineMapType_MapValues(x0=x0,
                                         m0=m0)

    # Count duplicates
    mapType = determineMapType(x0=x0, x1=x1)

    return mapType
